"""
For the time being this will not be implemented as there is no helm support
"""
# from typing import List, Optional
#
# from avionix.kube.base_objects import Authorization
# from avionix.kube.meta import ObjectMeta
# from avionix.yaml.yaml_handling import HelmYaml
#
#
# class ResourceAttributes(HelmYaml):
#     """
#     :param name: Name is the name of the resource being requested for a "get" or \
#         deleted for a "delete". "" (empty) means all.
#     :param group: Group is the API Group of the Resource.  "*" means all.
#     :param resource: Resource is one of the existing resource types.  "*" means all.
#     :param subresource: Subresource is one of the existing resource types.  "" means \
#         none.
#     :param verb: Verb is a kubernetes resource API verb, like: get, list, watch, \
#         create, update, delete, proxy.  "*" means all.
#     :param version: Version is the API Version of the Resource.  "*" means all.
#     :param namespace: Namespace is the namespace of the action being requested.  \
#         Currently, there is no distinction between no namespace and all namespaces "" \
#         (empty) is defaulted for LocalSubjectAccessReviews "" (empty) is empty for \
#         cluster-scoped resources "" (empty) means "all" for namespace scoped resources \
#         from a SubjectAccessReview or SelfSubjectAccessReview
#     """
#
#     def __init__(
#         self,
#         name: str,
#         group: str,
#         resource: str,
#         subresource: str,
#         verb: str,
#         version: str,
#         namespace: Optional[str] = None,
#     ):
#         self.name = name
#         self.group = group
#         self.resource = resource
#         self.subresource = subresource
#         self.verb = verb
#         self.version = version
#         self.namespace = namespace
#
#
# class NonResourceAttributes(HelmYaml):
#     """
#     :param path: Path is the URL path of the request
#     :param verb: Verb is the standard HTTP verb
#     """
#
#     def __init__(self, path: str, verb: str):
#         self.path = path
#         self.verb = verb
#
#
# class SelfSubjectAccessReviewSpec(HelmYaml):
#     """
#     :param non_resource_attributes: NonResourceAttributes describes information for a \
#         non-resource access request
#     :param resource_attributes: ResourceAuthorizationAttributes describes information \
#         for a resource access request
#     """
#
#     def __init__(
#         self,
#         non_resource_attributes: NonResourceAttributes,
#         resource_attributes: ResourceAttributes,
#     ):
#         self.nonResourceAttributes = non_resource_attributes
#         self.resourceAttributes = resource_attributes
#
#
# class SelfSubjectAccessReview(Authorization):
#     """
#     :param metadata: None
#     :param spec: Spec holds information about the request being evaluated.  user and \
#         groups must be empty
#     :param api_version: APIVersion defines the versioned schema of this representation \
#         of an object. Servers should convert recognized schemas to the latest internal \
#         value, and may reject unrecognized values. More info: \
#         https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
#     """
#
#     def __init__(
#         self,
#         metadata: ObjectMeta,
#         spec: SelfSubjectAccessReviewSpec,
#         api_version: Optional[str] = None,
#     ):
#         super().__init__(api_version)
#         self.metadata = metadata
#         self.spec = spec
#
#
# class NonResourceRule(HelmYaml):
#     """
#     :param non_resource_urls: NonResourceURLs is a set of partial urls that a user \
#         should have access to.  *s are allowed, but only as the full, final step in \
#         the path.  "*" means all.
#     :param verbs: Verb is a list of kubernetes non-resource API verbs, like: get, post, \
#         put, delete, patch, head, options.  "*" means all.
#     """
#
#     def __init__(self, non_resource_urls: List[str], verbs: List[str]):
#         self.nonResourceURLs = non_resource_urls
#         self.verbs = verbs
#
#
# class ResourceRule(HelmYaml):
#     """
#     :param api_groups: APIGroups is the name of the APIGroup that contains the \
#         resources.  If multiple API groups are specified, any action requested against \
#         one of the enumerated resources in any API group will be allowed.  "*" means \
#         all.
#     :param resource_names: ResourceNames is an optional white list of names that the \
#         rule applies to.  An empty set means that everything is allowed.  "*" means \
#         all.
#     :param resources: Resources is a list of resources this rule applies to.  "*"
#         means all in the specified apiGroups.  "\*/foo" represents the subresource
#         'foo' for all resources in the specified apiGroups.
#     :param verbs: Verb is a list of kubernetes resource API verbs, like: get, list, \
#         watch, create, update, delete, proxy.  "*" means all.
#     """
#
#     def __init__(
#         self,
#         api_groups: List[str],
#         resource_names: List[str],
#         resources: List[str],
#         verbs: List[str],
#     ):
#         self.apiGroups = api_groups
#         self.resourceNames = resource_names
#         self.resources = resources
#         self.verbs = verbs
#
#
# class SubjectAccessReviewSpec(HelmYaml):
#     """
#     :param extra: Extra corresponds to the user.Info.GetExtra() method from the \
#         authenticator.  Since that is input to the authorizer it needs a reflection \
#         here.
#     :param groups: Groups is the groups you're testing for.
#     :param non_resource_attributes: NonResourceAttributes describes information for a \
#         non-resource access request
#     :param resource_attributes: ResourceAuthorizationAttributes describes information \
#         for a resource access request
#     :param user: User is the user you're testing for. If you specify "User" but not \
#         "Groups", then is it interpreted as "What if User were not a member of any \
#         groups
#     :param uid: UID information about the requesting user.
#     """
#
#     def __init__(
#         self,
#         extra: dict,
#         groups: List[str],
#         non_resource_attributes: NonResourceAttributes,
#         resource_attributes: ResourceAttributes,
#         user: str,
#         uid: Optional[str] = None,
#     ):
#         self.extra = extra
#         self.groups = groups
#         self.nonResourceAttributes = non_resource_attributes
#         self.resourceAttributes = resource_attributes
#         self.user = user
#         self.uid = uid
#
#
# class LocalSubjectAccessReview(Authorization):
#     """
#     :param metadata: None
#     :param spec: Spec holds information about the request being evaluated.  \
#         spec.namespace must be equal to the namespace you made the request against.  \
#         If empty, it is defaulted.
#     :param api_version: APIVersion defines the versioned schema of this representation \
#         of an object. Servers should convert recognized schemas to the latest internal \
#         value, and may reject unrecognized values. More info: \
#         https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
#     """
#
#     def __init__(
#         self,
#         metadata: ObjectMeta,
#         spec: SubjectAccessReviewSpec,
#         api_version: Optional[str] = None,
#     ):
#         super().__init__(api_version)
#         self.metadata = metadata
#         self.spec = spec
#
#
# class SelfSubjectRulesReviewSpec(HelmYaml):
#     """
#     :param namespace: Namespace to evaluate rules for. Required.
#     """
#
#     def __init__(self, namespace: Optional[str] = None):
#         self.namespace = namespace
#
#
# class SelfSubjectRulesReview(Authorization):
#     """
#     :param metadata: None
#     :param spec: Spec holds information about the request being evaluated.
#     :param api_version: APIVersion defines the versioned schema of this representation \
#         of an object. Servers should convert recognized schemas to the latest internal \
#         value, and may reject unrecognized values. More info: \
#         https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
#     """
#
#     def __init__(
#         self,
#         metadata: ObjectMeta,
#         spec: SelfSubjectRulesReviewSpec,
#         api_version: Optional[str] = None,
#     ):
#         super().__init__(api_version)
#         self.metadata = metadata
#         self.spec = spec
#
#
# class SubjectAccessReview(Authorization):
#     """
#     :param metadata: None
#     :param spec: Spec holds information about the request being evaluated
#     :param api_version: APIVersion defines the versioned schema of this representation \
#         of an object. Servers should convert recognized schemas to the latest internal \
#         value, and may reject unrecognized values. More info: \
#         https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
#     """
#
#     def __init__(
#         self,
#         metadata: ObjectMeta,
#         spec: SubjectAccessReviewSpec,
#         api_version: Optional[str] = None,
#     ):
#         super().__init__(api_version)
#         self.metadata = metadata
#         self.spec = spec



"""
Classes related to Deployments, StatefulSets, ReplicaSets and DaemonSets
"""

from typing import List, Optional, Union

from avionix.kube.base_objects import Apps
from avionix.kube.core import PersistentVolumeClaim, PodTemplateSpec
from avionix.kube.meta import LabelSelector, ObjectMeta
from avionix.yaml.yaml_handling import HelmYaml


class RollingUpdateStatefulSetStrategy(HelmYaml):
    """
    :param partition: Partition indicates the ordinal at which the StatefulSet should \
        be partitioned. Default value is 0.
    """

    def __init__(self, partition: int):
        self.partition = partition


class StatefulSetUpdateStrategy(HelmYaml):
    """
    :param rolling_update: RollingUpdate is used to communicate parameters when Type is \
        RollingUpdateStatefulSetStrategyType.
    :param type: Type indicates the type of the StatefulSetUpdateStrategy. Default is \
        RollingUpdate.
    """

    def __init__(
        self,
        rolling_update: Optional[RollingUpdateStatefulSetStrategy] = None,
        type: Optional[str] = None,
    ):
        self.rollingUpdate = rolling_update
        self.type = type


class StatefulSetSpec(HelmYaml):
    """
    :param template: template is the object that describes the pod that will be created \
        if insufficient replicas are detected. Each pod stamped out by the StatefulSet \
        will fulfill this Template, but have a unique identity from the rest of the \
        StatefulSet.
    :param selector: selector is a label query over pods that should match the replica \
        count. It must match the pod template's labels. More info: \
        https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#label-selectors  # noqa
    :param service_name: serviceName is the name of the service that governs this \
        StatefulSet. This service must exist before the StatefulSet, and is \
        responsible for the network identity of the set. Pods get DNS/hostnames that \
        follow the pattern: pod-specific-string.serviceName.default.svc.cluster.local \
        where "pod-specific-string" is managed by the StatefulSet controller.
    :param pod_management_policy: podManagementPolicy controls how pods are created \
        during initial scale up, when replacing pods on nodes, or when scaling down. \
        The default policy is `OrderedReady`, where pods are created in increasing \
        order (pod-0, then pod-1, etc) and the controller will wait until each pod is \
        ready before continuing. When scaling down, the pods are removed in the \
        opposite order. The alternative policy is `Parallel` which will create pods in \
        parallel to match the desired scale without waiting, and on scale down will \
        delete all pods at once.
    :param revision_history_limit: revisionHistoryLimit is the maximum number of \
        revisions that will be maintained in the StatefulSet's revision history. The \
        revision history consists of all revisions not represented by a currently \
        applied StatefulSetSpec version. The default value is 10.
    :param volume_claim_templates: volumeClaimTemplates is a list of claims that pods \
        are allowed to reference. The StatefulSet controller is responsible for \
        mapping network identities to claims in a way that maintains the identity of a \
        pod. Every claim in this list must have at least one matching (by name) \
        volumeMount in one container in the template. A claim in this list takes \
        precedence over any volumes in the template, with the same name.
    :param replicas: replicas is the desired number of replicas of the given Template. \
        These are replicas in the sense that they are instantiations of the same \
        Template, but individual replicas also have a consistent identity. If \
        unspecified, defaults to 1.
    :param update_strategy: updateStrategy indicates the StatefulSetUpdateStrategy that \
        will be employed to update Pods in the StatefulSet when a revision is made to \
        Template.
    """

    def __init__(
        self,
        template: PodTemplateSpec,
        selector: LabelSelector,
        service_name: str,
        pod_management_policy: Optional[str] = None,
        revision_history_limit: Optional[int] = None,
        volume_claim_templates: Optional[List[PersistentVolumeClaim]] = None,
        replicas: Optional[int] = None,
        update_strategy: Optional[StatefulSetUpdateStrategy] = None,
    ):
        self.template = template
        self.podManagementPolicy = pod_management_policy
        self.revisionHistoryLimit = revision_history_limit
        self.selector = selector
        self.serviceName = service_name
        self.volumeClaimTemplates = volume_claim_templates
        self.replicas = replicas
        self.updateStrategy = update_strategy


class StatefulSet(Apps):
    """
    :param metadata: None
    :param spec: Spec defines the desired identities of pods in this set.
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        spec: StatefulSetSpec,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec


class ControllerRevision(Apps):
    """
    :param metadata: Standard object's metadata. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata  # noqa
    :param data: Data is the serialized representation of the state.
    :param revision: Revision indicates the revision of the state represented by Data.
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        data: dict,
        revision: int,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.data = data
        self.revision = revision


class RollingUpdateDeployment(HelmYaml):
    """
    :param max_surge: The maximum number of pods that can be scheduled above the \
        desired number of pods. Value can be an absolute number (ex: 5) or a \
        percentage of desired pods (ex: 10%). This can not be 0 if MaxUnavailable is \
        0. Absolute number is calculated from percentage by rounding up. Defaults to \
        25%. Example: when this is set to 30%, the new ReplicaSet can be scaled up \
        immediately when the rolling update starts, such that the total number of old \
        and new pods do not exceed 130% of desired pods. Once old pods have been \
        killed, new ReplicaSet can be scaled up further, ensuring that total number of \
        pods running at any time during the update is at most 130% of desired pods.
    :param max_unavailable: The maximum number of pods that can be unavailable during \
        the update. Value can be an absolute number (ex: 5) or a percentage of desired \
        pods (ex: 10%). Absolute number is calculated from percentage by rounding \
        down. This can not be 0 if MaxSurge is 0. Defaults to 25%. Example: when this \
        is set to 30%, the old ReplicaSet can be scaled down to 70% of desired pods \
        immediately when the rolling update starts. Once new pods are ready, old \
        ReplicaSet can be scaled down further, followed by scaling up the new \
        ReplicaSet, ensuring that the total number of pods available at all times \
        during the update is at least 70% of desired pods.
    """

    def __init__(
        self,
        max_surge: Optional[Union[int, str]] = None,
        max_unavailable: Optional[Union[int, str]] = None,
    ):
        self.maxSurge = max_surge
        self.maxUnavailable = max_unavailable


class RollingUpdateDaemonSet(HelmYaml):
    """
    :param max_unavailable: The maximum number of DaemonSet pods that can be \
        unavailable during the update. Value can be an absolute number (ex: 5) or a \
        percentage of total number of DaemonSet pods at the start of the update (ex: \
        10%). Absolute number is calculated from percentage by rounding up. This \
        cannot be 0. Default value is 1. Example: when this is set to 30%, at most 30% \
        of the total number of nodes that should be running the daemon pod (i.e. \
        status.desiredNumberScheduled) can have their pods stopped for an update at \
        any given time. The update starts by stopping at most 30% of those DaemonSet \
        pods and then brings up new DaemonSet pods in their place. Once the new pods \
        are available, it then proceeds onto other DaemonSet pods, thus ensuring that \
        at least 70% of original number of DaemonSet pods are available at all times \
        during the update.
    """

    def __init__(self, max_unavailable: Union[int, str]):
        self.maxUnavailable = max_unavailable


class DaemonSetUpdateStrategy(HelmYaml):
    """
    :param rolling_update: Rolling update config params. Present only if type = \
        "RollingUpdate".
    :param type: Type of daemon set update. Can be "RollingUpdate" or "OnDelete". \
        Default is RollingUpdate.
    """

    def __init__(
        self,
        rolling_update: Optional[RollingUpdateDaemonSet] = None,
        type: Optional[str] = None,
    ):
        self.rollingUpdate = rolling_update
        self.type = type


class DaemonSetSpec(HelmYaml):
    """
    :param template: An object that describes the pod that will be created. The \
        DaemonSet will create exactly one copy of this pod on every node that matches \
        the template's node selector (or on every node if no node selector is \
        specified). More info: \
        https://kubernetes.io/docs/concepts/workloads/controllers/replicationcontroller#pod-template  # noqa
    :param selector: A label query over pods that are managed by the daemon set. Must \
        match in order to be controlled. It must match the pod template's labels. More \
        info: \
        https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#label-selectors  # noqa
    :param min_ready_seconds: The minimum number of seconds for which a newly created \
        DaemonSet pod should be ready without any of its container crashing, for it to \
        be considered available. Defaults to 0 (pod will be considered available as \
        soon as it is ready).
    :param revision_history_limit: The number of old history to retain to allow \
        rollback. This is a pointer to distinguish between explicit zero and not \
        specified. Defaults to 10.
    :param update_strategy: An update strategy to replace existing DaemonSet pods with \
        new pods.
    """

    def __init__(
        self,
        template: PodTemplateSpec,
        selector: LabelSelector,
        min_ready_seconds: Optional[int] = None,
        revision_history_limit: Optional[int] = None,
        update_strategy: Optional[DaemonSetUpdateStrategy] = None,
    ):
        self.template = template
        self.selector = selector
        self.minReadySeconds = min_ready_seconds
        self.revisionHistoryLimit = revision_history_limit
        self.updateStrategy = update_strategy


class DaemonSet(Apps):
    """
    :param metadata: Standard object's metadata. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata  # noqa
    :param spec: The desired behavior of this daemon set. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#spec-and-status  # noqa
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        spec: DaemonSetSpec,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec


class ReplicaSetSpec(HelmYaml):
    """
    :param template: Template is the object that describes the pod that will be created \
        if insufficient replicas are detected. More info: \
        https://kubernetes.io/docs/concepts/workloads/controllers/replicationcontroller#pod-template  # noqa
    :param selector: Selector is a label query over pods that should match the replica \
        count. Label keys and values that must match in order to be controlled by this \
        replica set. It must match the pod template's labels. More info: \
        https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#label-selectors  # noqa
    :param min_ready_seconds: Minimum number of seconds for which a newly created pod \
        should be ready without any of its container crashing, for it to be considered \
        available. Defaults to 0 (pod will be considered available as soon as it is \
        ready)
    :param replicas: Replicas is the number of desired replicas. This is a pointer to \
        distinguish between explicit zero and unspecified. Defaults to 1. More info: \
        https://kubernetes.io/docs/concepts/workloads/controllers/replicationcontroller/#what-is-a-replicationcontroller  # noqa
    """

    def __init__(
        self,
        template: PodTemplateSpec,
        selector: LabelSelector,
        min_ready_seconds: Optional[int] = None,
        replicas: Optional[int] = None,
    ):
        self.template = template
        self.selector = selector
        self.minReadySeconds = min_ready_seconds
        self.replicas = replicas


class ReplicaSet(Apps):
    """
    :param metadata: If the Labels of a ReplicaSet are empty, they are defaulted to be \
        the same as the Pod(s) that the ReplicaSet manages. Standard object's \
        metadata. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata  # noqa
    :param spec: Spec defines the specification of the desired behavior of the \
        ReplicaSet. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#spec-and-status  # noqa
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        spec: ReplicaSetSpec,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec


class DeploymentStrategy(HelmYaml):
    """
    :param rolling_update: Rolling update config params. Present only if \
        DeploymentStrategyType = RollingUpdate.
    :param type: Type of deployment. Can be "Recreate" or "RollingUpdate". Default is \
        RollingUpdate.
    """

    def __init__(
        self,
        rolling_update: Optional[RollingUpdateDeployment] = None,
        type: Optional[str] = None,
    ):
        self.rollingUpdate = rolling_update
        self.type = type


class DeploymentSpec(HelmYaml):
    """
    :param template: Template describes the pods that will be created.
    :param selector: Label selector for pods. Existing ReplicaSets whose pods are \
        selected by this will be the ones affected by this deployment. It must match \
        the pod template's labels.
    :param min_ready_seconds: Minimum number of seconds for which a newly created pod \
        should be ready without any of its container crashing, for it to be considered \
        available. Defaults to 0 (pod will be considered available as soon as it is \
        ready)
    :param paused: Indicates that the deployment is paused.
    :param progress_deadline_seconds: The maximum time in seconds for a deployment to \
        make progress before it is considered to be failed. The deployment controller \
        will continue to process failed deployments and a condition with a \
        ProgressDeadlineExceeded reason will be surfaced in the deployment status. \
        Note that progress will not be estimated during the time a deployment is \
        paused. Defaults to 600s.
    :param replicas: Number of desired pods. This is a pointer to distinguish between \
        explicit zero and not specified. Defaults to 1.
    :param revision_history_limit: The number of old ReplicaSets to retain to allow \
        rollback. This is a pointer to distinguish between explicit zero and not \
        specified. Defaults to 10.
    :param strategy: The deployment strategy to use to replace existing pods with new \
        ones.
    """

    def __init__(
        self,
        template: PodTemplateSpec,
        selector: LabelSelector,
        min_ready_seconds: Optional[int] = None,
        paused: Optional[bool] = None,
        progress_deadline_seconds: Optional[int] = None,
        replicas: Optional[int] = None,
        revision_history_limit: Optional[int] = None,
        strategy: Optional[DeploymentStrategy] = None,
    ):
        self.template = template
        self.selector = selector
        self.minReadySeconds = min_ready_seconds
        self.paused = paused
        self.progressDeadlineSeconds = progress_deadline_seconds
        self.replicas = replicas
        self.revisionHistoryLimit = revision_history_limit
        self.strategy = strategy


class Deployment(Apps):
    """
    :param metadata: Standard object metadata.
    :param spec: Specification of the desired behavior of the Deployment.
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        spec: DeploymentSpec,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec



from typing import Optional

from avionix._process_utils import custom_check_output
from avionix.testing.helpers import parse_output_to_dict


def get_helm_installations(namespace: Optional[str] = None):
    command = "helm list"
    if namespace is not None:
        command += f" -n {namespace}"
    output = custom_check_output(command)
    return parse_output_to_dict(output)



"""
Classes related to networking (NetworkPolicy and IngressClass)
"""

from typing import List, Optional

from avionix.kube.base_objects import Networking
from avionix.kube.core import TypedLocalObjectReference
from avionix.kube.meta import LabelSelector, ObjectMeta
from avionix.yaml.yaml_handling import HelmYaml


class IPBlock(HelmYaml):
    """
    :param cidr: CIDR is a string representing the IP Block Valid examples are \
        "192.168.1.1/24" or "2001:db9::/64"
    :param except_: Except is a slice of CIDRs that should not be included within an IP \
        Block Valid examples are "192.168.1.1/24" or "2001:db9::/64" Except values \
        will be rejected if they are outside the CIDR range
    """

    def __init__(self, cidr: str, except_: Optional[List[str]] = None):
        self.cidr = cidr
        self["except"] = except_


class NetworkPolicyPeer(HelmYaml):
    """
    :param ip_block: IPBlock defines policy on a particular IPBlock. If this field is \
        set then neither of the other fields can be.
    :param namespace_selector: Selects Namespaces using cluster-scoped labels. This \
        field follows standard label selector semantics; if present but empty, it \
        selects all namespaces.  If PodSelector is also set, then the \
        NetworkPolicyPeer as a whole selects the Pods matching PodSelector in the \
        Namespaces selected by NamespaceSelector. Otherwise it selects all Pods in the \
        Namespaces selected by NamespaceSelector.
    :param pod_selector: This is a label selector which selects Pods. This field \
        follows standard label selector semantics; if present but empty, it selects \
        all pods.  If NamespaceSelector is also set, then the NetworkPolicyPeer as a \
        whole selects the Pods matching PodSelector in the Namespaces selected by \
        NamespaceSelector. Otherwise it selects the Pods matching PodSelector in the \
        policy's own Namespace.
    """

    def __init__(
        self,
        ip_block: IPBlock,
        namespace_selector: Optional[LabelSelector] = None,
        pod_selector: Optional[LabelSelector] = None,
    ):
        self.ipBlock = ip_block
        self.namespaceSelector = namespace_selector
        self.podSelector = pod_selector


class NetworkPolicyPort(HelmYaml):
    """
    :param port: The port on the given protocol. This can either be a numerical or \
        named port on a pod. If this field is not provided, this matches all port \
        names and numbers.
    :param protocol: The protocol (TCP, UDP, or SCTP) which traffic must match. If not \
        specified, this field defaults to TCP.
    """

    def __init__(self, port: Optional[int] = None, protocol: Optional[str] = None):
        self.port = port
        self.protocol = protocol


class NetworkPolicyIngressRule(HelmYaml):
    """
    :param from_: List of sources which should be able to access the pods selected for \
        this rule. Items in this list are combined using a logical OR operation. If \
        this field is empty or missing, this rule matches all sources (traffic not \
        restricted by source). If this field is present and contains at least one \
        item, this rule allows traffic only if the traffic matches at least one item \
        in the from list.
    :param ports: List of ports which should be made accessible on the pods selected \
        for this rule. Each item in this list is combined using a logical OR. If this \
        field is empty or missing, this rule matches all ports (traffic not restricted \
        by port). If this field is present and contains at least one item, then this \
        rule allows traffic only if the traffic matches at least one port in the list.
    """

    def __init__(self, from_: List[NetworkPolicyPeer], ports: List[NetworkPolicyPort]):
        self["from"] = from_
        self.ports = ports


class NetworkPolicyEgressRule(HelmYaml):
    """
    :param ports: List of destination ports for outgoing traffic. Each item in this \
        list is combined using a logical OR. If this field is empty or missing, this \
        rule matches all ports (traffic not restricted by port). If this field is \
        present and contains at least one item, then this rule allows traffic only if \
        the traffic matches at least one port in the list.
    :param to: List of destinations for outgoing traffic of pods selected for this \
        rule. Items in this list are combined using a logical OR operation. If this \
        field is empty or missing, this rule matches all destinations (traffic not \
        restricted by destination). If this field is present and contains at least one \
        item, this rule allows traffic only if the traffic matches at least one item \
        in the to list.
    """

    def __init__(self, ports: List[NetworkPolicyPort], to: List[NetworkPolicyPeer]):
        self.ports = ports
        self.to = to


class NetworkPolicySpec(HelmYaml):
    """
    :param egress: List of egress rules to be applied to the selected pods. Outgoing \
        traffic is allowed if there are no NetworkPolicies selecting the pod (and \
        cluster policy otherwise allows the traffic), OR if the traffic matches at \
        least one egress rule across all of the NetworkPolicy objects whose \
        podSelector matches the pod. If this field is empty then this NetworkPolicy \
        limits all outgoing traffic (and serves solely to ensure that the pods it \
        selects are isolated by default). This field is beta-level in 1.8
    :param ingress: List of ingress rules to be applied to the selected pods. Traffic \
        is allowed to a pod if there are no NetworkPolicies selecting the pod (and \
        cluster policy otherwise allows the traffic), OR if the traffic source is the \
        pod's local node, OR if the traffic matches at least one ingress rule across \
        all of the NetworkPolicy objects whose podSelector matches the pod. If this \
        field is empty then this NetworkPolicy does not allow any traffic (and serves \
        solely to ensure that the pods it selects are isolated by default)
    :param pod_selector: Selects the pods to which this NetworkPolicy object applies. \
        The array of ingress rules is applied to any pods selected by this field. \
        Multiple network policies can select the same set of pods. In this case, the \
        ingress rules for each are combined additively. This field is NOT optional and \
        follows standard label selector semantics. An empty podSelector matches all \
        pods in this namespace.
    :param policy_types: List of rule types that the NetworkPolicy relates to. Valid \
        options are "Ingress", "Egress", or "Ingress,Egress". If this field is not \
        specified, it will default based on the existence of Ingress or Egress rules; \
        policies that contain an Egress section are assumed to affect Egress, and all \
        policies (whether or not they contain an Ingress section) are assumed to \
        affect Ingress. If you want to write an egress-only policy, you must \
        explicitly specify policyTypes [ "Egress" ]. Likewise, if you want to write a \
        policy that specifies that no egress is allowed, you must specify a \
        policyTypes value that include "Egress" (since such a policy would not include \
        an Egress section and would otherwise default to just [ "Ingress" ]). This \
        field is beta-level in 1.8
    """

    def __init__(
        self,
        egress: Optional[List[NetworkPolicyEgressRule]],
        ingress: Optional[List[NetworkPolicyIngressRule]],
        pod_selector: Optional[LabelSelector],
        policy_types: Optional[List[str]],
    ):
        self.egress = egress
        self.ingress = ingress
        self.podSelector = pod_selector
        self.policyTypes = policy_types


class NetworkPolicy(Networking):
    """
    :param metadata: Standard object's metadata. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata  # noqa
    :param spec: Specification of the desired behavior for this NetworkPolicy.
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        spec: NetworkPolicySpec,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec


class IngressClassSpec(HelmYaml):
    """
    :param controller: Controller refers to the name of the controller that should \
        handle this class. This allows for different "flavors" that are controlled by \
        the same controller. For example, you may have different Parameters for the \
        same implementing controller. This should be specified as a domain-prefixed \
        path no more than 250 characters in length, e.g. "acme.io/ingress-controller". \
        This field is immutable.
    :param parameters: Parameters is a link to a custom resource containing additional \
        configuration for the controller. This is optional if the controller does not \
        require extra parameters.
    """

    def __init__(
        self, controller: str, parameters: Optional[TypedLocalObjectReference] = None
    ):
        self.controller = controller
        self.parameters = parameters


class IngressClass(Networking):
    """
    :param metadata: Standard object's metadata. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata  # noqa
    :param spec: Spec is the desired state of the IngressClass. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#spec-and-status  # noqa
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    _non_standard_version = "v1beta1"

    def __init__(
        self,
        metadata: ObjectMeta,
        spec: IngressClassSpec,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec



"""
Classes related to Ingress
"""

from typing import List, Optional

from avionix.kube.base_objects import Extensions
from avionix.kube.core import TypedLocalObjectReference
from avionix.kube.meta import ObjectMeta
from avionix.yaml.yaml_handling import HelmYaml


class IngressTLS(HelmYaml):
    """
    :param secret_name: SecretName is the name of the secret used to terminate SSL \
        traffic on 443. Field is left optional to allow SSL routing based on SNI \
        hostname alone. If the SNI host in a listener conflicts with the "Host" header \
        field used by an IngressRule, the SNI host is used for termination and value \
        of the Host header is used for routing.
    :param hosts: Hosts are a list of hosts included in the TLS certificate. The values \
        in this list must match the name/s used in the tlsSecret. Defaults to the \
        wildcard host setting for the loadbalancer controller fulfilling this Ingress, \
        if left unspecified.
    """

    def __init__(self, secret_name: str, hosts: Optional[List[str]] = None):
        self.secretName = secret_name
        self.hosts = hosts


class IngressBackend(HelmYaml):
    """
    :param service_name: Specifies the name of the referenced service.
    :param service_port: Specifies the port of the referenced service.
    :param resource: Resource is an ObjectRef to another Kubernetes resource in the \
        namespace of the Ingress object. If resource is specified, serviceName and \
        servicePort must not be specified.
    """

    def __init__(
        self,
        service_name: str,
        service_port: int,
        resource: Optional[TypedLocalObjectReference] = None,
    ):
        self.resource = resource
        self.serviceName = service_name
        self.servicePort = service_port


class HTTPIngressPath(HelmYaml):
    """
    :param backend: Backend defines the referenced service endpoint to which the \
        traffic will be forwarded to.
    :param path: Path is matched against the path of an incoming request. Currently it \
        can contain characters disallowed from the conventional "path" part of a URL \
        as defined by RFC 3986. Paths must begin with a '/'. When unspecified, all \
        paths from incoming requests are matched.
    :param path_type: PathType determines the interpretation of the Path matching. \
        PathType can be one of the following values: * Exact: Matches the URL path \
        exactly. * Prefix: Matches based on a URL path prefix split by '/'. Matching \
        is   done on a path element by element basis. A path element refers is the   \
        list of labels in the path split by the '/' separator. A request is a   match \
        for path p if every p is an element-wise prefix of p of the   request path. \
        Note that if the last element of the path is a substring   of the last element \
        in request path, it is not a match (e.g. /foo/bar   matches /foo/bar/baz, but \
        does not match /foo/barbaz). * ImplementationSpecific: Interpretation of the \
        Path matching is up to   the IngressClass. Implementations can treat this as a \
        separate PathType   or treat it identically to Prefix or Exact path types. \
        Implementations are required to support all path types. Defaults to \
        ImplementationSpecific.
    """

    def __init__(
        self,
        backend: IngressBackend,
        path: Optional[str] = None,
        path_type: Optional[str] = None,
    ):
        self.backend = backend
        self.path = path
        self.pathType = path_type


class HTTPIngressRuleValue(HelmYaml):
    """
    :param paths: A collection of paths that map requests to backends.
    """

    def __init__(self, paths: List[HTTPIngressPath]):
        self.paths = paths


class IngressRule(HelmYaml):
    """
    :param http: None
    :param host: Host is the fully qualified domain name of a network host, as defined \
        by RFC 3986. Note the following deviations from the "host" part of the URI as \
        defined in RFC 3986: 1. IPs are not allowed. Currently an IngressRuleValue can \
        only apply to    the IP in the Spec of the parent Ingress. 2. The `:` \
        delimiter is not respected because ports are not allowed.       Currently the \
        port of an Ingress is implicitly :80 for http and       :443 for https. Both \
        these may change in the future. Incoming requests are matched against the host \
        before the IngressRuleValue. If the host is unspecified, the Ingress routes \
        all traffic based on the specified IngressRuleValue.  Host can be "precise" \
        which is a domain name without the terminating dot of a network host (e.g. \
        "foo.bar.com") or "wildcard", which is a domain name prefixed with a single \
        wildcard label (e.g. "*.foo.com"). The wildcard character '\\*' must appear by \
        itself as the first DNS label and matches only a single label. You cannot have \
        a wildcard label by itself (e.g. Host == "*"). Requests will be matched \
        against the Host field in the following way: 1. If Host is precise, the \
        request matches this rule if the http host header is equal to Host. 2. If Host \
        is a wildcard, then the request matches this rule if the http host header is \
        to equal to the suffix (removing the first label) of the wildcard rule.
    """

    def __init__(self, http: HTTPIngressRuleValue, host: Optional[str] = None):
        self.host = host
        self.http = http


class IngressSpec(HelmYaml):
    """
    :param ingress_class_name: IngressClassName is the name of the IngressClass cluster \
        resource. The associated IngressClass defines which controller will implement \
        the resource. This replaces the deprecated `kubernetes.io/ingress.class` \
        annotation. For backwards compatibility, when that annotation is set, it must \
        be given precedence over this field. The controller may emit a warning if the \
        field and annotation have different values. Implementations of this API should \
        ignore Ingresses without a class specified. An IngressClass resource may be \
        marked as default, which can be used to set a default value for this field. \
        For more information, refer to the IngressClass documentation.
    :param backend: A default backend capable of servicing requests that don't match \
        any rule. At least one of 'backend' or 'rules' must be specified. This field \
        is optional to allow the loadbalancer controller or defaulting logic to \
        specify a global default.
    :param rules: A list of host rules used to configure the Ingress. If unspecified, \
        or no rule matches, all traffic is sent to the default backend.
    :param tls: TLS configuration. Currently the Ingress only supports a single TLS \
        port, 443. If multiple members of this list specify different hosts, they will \
        be multiplexed on the same port according to the hostname specified through \
        the SNI TLS extension, if the ingress controller fulfilling the ingress \
        supports SNI.
    """

    def __init__(
        self,
        ingress_class_name: Optional[str] = None,
        backend: Optional[IngressBackend] = None,
        rules: Optional[List[IngressRule]] = None,
        tls: Optional[List[IngressTLS]] = None,
    ):
        self.ingressClassName = ingress_class_name
        self.backend = backend
        self.rules = rules
        self.tls = tls


class Ingress(Extensions):
    """
    :param metadata: Standard object's metadata. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata  # noqa
    :param spec: Spec is the desired state of the Ingress. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#spec-and-status  # noqa
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    _non_standard_version = "v1beta1"

    def __init__(
        self, metadata: ObjectMeta, spec: IngressSpec, api_version: Optional[str] = None
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec



from logging import error, info
import os
from pathlib import Path
import re
import shutil
import subprocess
from typing import Dict, List, Optional

import yaml

from avionix._process_utils import custom_check_output
from avionix.chart.chart_info import ChartInfo
from avionix.chart.utils import get_helm_installations
from avionix.chart.values_yaml import Values
from avionix.errors import (
    ChartNotInstalledError,
    ErrorFactory,
    post_uninstall_handle_error,
)
from avionix.kube.base_objects import KubernetesBaseObject


class ChartBuilder:
    """
    Main builder object. Accepts kubernetes objects and generates the helm chart
    structure. Can also perform the installation onto the server

    :param chart_info: Contains all chart metadata and dependency info
    :param kubernetes_objects: A list of kubernetes objects
    :param output_directory: A path to the directory in which to place the generated \
        chart
    :param keep_chart: Whether or not to keep the chart after installation
    :param namespace: The namespace in which all chart components should be installed \
        This allows the convenience of not passing the namespace option to both \
        install and uninstall
    """

    def __init__(
        self,
        chart_info: ChartInfo,
        kubernetes_objects: List[KubernetesBaseObject],
        output_directory: Optional[str] = None,
        keep_chart: bool = False,
        namespace: Optional[str] = None,
        values: Optional[Values] = None,
    ):
        self.chart_info = chart_info
        self.kubernetes_objects = kubernetes_objects
        self.chart_folder_path = Path(self.chart_info.name)
        self.__templates_directory = self.chart_folder_path / "templates"
        self.__chart_yaml = self.chart_folder_path / "Chart.yaml"
        self.__keep_chart = keep_chart
        self.__values = values
        self.namespace = namespace
        if output_directory:
            self.__templates_directory = Path(output_directory) / str(
                self.__templates_directory
            )
            self.__chart_yaml = Path(output_directory) / str(self.__chart_yaml)
            self.chart_folder_path = Path(output_directory) / str(
                self.chart_folder_path
            )

    def __delete_chart_directory(self):
        if os.path.exists(self.chart_info.name):
            shutil.rmtree(self.chart_info.name)

    def generate_chart(self):
        """
        Generates the chart but does not install it on kubernetes

        :returns The template directory
        """
        self.__delete_chart_directory()
        os.makedirs(self.__templates_directory, exist_ok=True)
        with open(self.__chart_yaml, "w+") as chart_yaml_file:
            chart_yaml_file.write(str(self.chart_info))

        kind_count: Dict[str, int] = {}
        for kubernetes_object in self.kubernetes_objects:
            if kubernetes_object.kind not in kind_count:
                kind_count[kubernetes_object.kind] = 0
            else:
                kind_count[kubernetes_object.kind] += 1
            with open(
                self.__templates_directory / f"{kubernetes_object.kind}-"
                f"{kind_count[kubernetes_object.kind]}.yaml",
                "w",
            ) as template:
                template.write(str(kubernetes_object))
        with open(
            self.__templates_directory.parent / "values.yaml", "w"
        ) as values_file:
            values_file.write(self.__get_values_yaml())
        return self.__templates_directory

    def _helm_list_repos(self) -> List[str]:
        try:
            return custom_check_output("helm repo list").split("\n")[1:]
        except subprocess.CalledProcessError as err:
            error_message = err.output.decode("utf-8").strip()
            if error_message == "Error: no repositories to show":
                return []
            error(error_message)
            raise err

    def get_helm_repos(self):
        repo_lines = self._helm_list_repos()
        repo_to_url_dict = {}
        for repo_line in repo_lines:
            repo_line_no_extra_space = repo_line.strip()
            match = re.match(
                "(?P<repo_name>.+?)\s+(?P<url>.+)", repo_line_no_extra_space
            )
            if not repo_line_no_extra_space:
                continue
            if not match:
                raise Exception(
                    f"Could not match repo name pattern from output for "
                    f"line {repo_line_no_extra_space}"
                )
            repo_to_url_dict[match.group("repo_name")] = match.group("url")
        return repo_to_url_dict

    def add_dependency_repos(self):
        """
        Adds repos for all dependencies listed
        """
        info("Adding dependencies...")
        installed_repos = self.get_helm_repos()
        for dependency in self.chart_info.dependencies:
            if installed_repos.get(dependency.local_repo_name) == dependency.repository:
                continue
            dependency.add_repo()

    def __get_values_yaml(self):
        values = {}
        for dependency in self.chart_info.dependencies:
            values.update(dependency.get_values_yaml())
        if self.__values:
            values.update(self.__values.values)
        return yaml.dump(values)

    @staticmethod
    def __parse_options(options: Optional[Dict[str, Optional[str]]] = None):
        option_string = ""
        if options is None:
            return option_string
        for option in options:
            option_string += f" --{option}"

            # Add value after flag if one is given
            value = options[option]
            if value:
                option_string += f" {value}"
        return option_string

    def __get_helm_install_command(
        self, options: Optional[Dict[str, Optional[str]]] = None
    ):
        command = (
            f"helm install {self.chart_info.name} {self.chart_folder_path.resolve()}"
        )
        return self.__handle_namespace(command) + self.__parse_options(options)

    def run_helm_install(self, options: Optional[Dict[str, Optional[str]]] = None):
        """
        Runs helm install on the chart

        :param options: A dictionary of command line arguments to pass to helm

        :Example:

        To run an install with updated dependencies and with verbose logging:

        >>> self.run_helm_install({"dependency_update": None, "v": "info"})
        """
        custom_check_output(self.__get_helm_install_command(options))

    def __handle_installation(self, options: Optional[Dict[str, Optional[str]]] = None):
        try:
            info(f"Installing helm chart {self.chart_info.name}...")
            self.run_helm_install(options)
        except subprocess.CalledProcessError as err:
            decoded = err.output.decode("utf-8")
            error = ErrorFactory(decoded).get_error()
            if error is not None:
                raise error
            if self.is_installed:
                self.uninstall_chart()
            raise post_uninstall_handle_error(decoded)

    def install_chart(self, options: Optional[Dict[str, Optional[str]]] = None):
        """
        Generates and installs the helm chart onto kubernetes and handles all failures.
        It will also add the repos of all listed dependencies.

        Note that the generated chart will be deleted if *keep_chart* is not set to
        true on ChartBuilder

        WARNING: If the helm chart installation fails, the chart will be uninstalled,
        so if working with an existing chart, please use upgrade_chart instead

        :param options: A dictionary of command line arguments to pass to helm

        For example, to run an install with updated dependencies and with verbose
        logging:
        >>> self.helm_install({"dependency_update": None, "v": "info"})
        """
        self.generate_chart()
        self.add_dependency_repos()
        self.__handle_installation(options)
        if not self.__keep_chart:
            self.__delete_chart_directory()

    def __get_helm_uninstall_command(
        self, options: Optional[Dict[str, Optional[str]]] = None
    ):
        command = f"helm uninstall {self.chart_info.name}"
        return self.__handle_namespace(command) + self.__parse_options(options)

    def run_helm_uninstall(self, options: Optional[Dict[str, Optional[str]]] = None):
        """
        Runs helm uninstall

        :param options: A dictionary of command line arguments to pass to helm

        :Example:

        >>> self.run_helm_uninstall(
        >>>    {"dry-run": None,
        >>>    "description": "My uninstall description"
        >>>    }
        >>> )
        """
        info(f"Uninstalling chart {self.chart_info.name}")
        custom_check_output(self.__get_helm_uninstall_command(options))

    def __check_if_installed(self):
        info(f"Checking if helm chart {self.chart_info.name} is installed")
        if not self.is_installed:
            raise ChartNotInstalledError(
                f'Error: chart "{self.chart_info.name}" is not installed'
            )

    def __handle_uninstallation(
        self, options: Optional[Dict[str, Optional[str]]] = None
    ):
        self.__check_if_installed()
        self.run_helm_uninstall(options)

    def uninstall_chart(self, options: Optional[Dict[str, Optional[str]]] = None):
        """
        Uninstalls the chart if present, if not present, raises an error

        :param options: A dictionary of command line arguments to pass to helm

        :Example:

        >>> self.uninstall_chart(
        >>>    {"dry-run": None,
        >>>    "description": "My uninstall description"
        >>>    }
        >>> )
        """
        self.__handle_uninstallation(options)

    def __handle_namespace(self, command: str):
        if self.namespace is not None:
            return command + f" -n {self.namespace}"
        return command

    def __get_helm_upgrade_command(
        self, options: Optional[Dict[str, Optional[str]]] = None
    ):
        command = f"helm upgrade {self.chart_info.name} {self.chart_folder_path}"
        return self.__handle_namespace(command) + self.__parse_options(options)

    def __handle_upgrade(self, options: Optional[Dict[str, Optional[str]]] = None):
        try:
            self.run_helm_upgrade(options)
        except subprocess.CalledProcessError as err:
            decoded = err.output.decode("utf-8")
            error = ErrorFactory(decoded).get_error()
            if error is not None:
                raise error
            raise post_uninstall_handle_error(decoded)

    def run_helm_upgrade(self, options: Optional[Dict[str, Optional[str]]] = None):
        """
        Runs 'helm upgrade' on the chart

        :param options: A dictionary of command line arguments to pass to helm

        :Example:

        >>> self.run_helm_upgrade(options={"atomic": None, "version": "2.0"})
        """
        info(f"Upgrading helm chart {self.chart_info.name}")
        custom_check_output(self.__get_helm_upgrade_command(options))

    def upgrade_chart(self, options: Optional[Dict[str, Optional[str]]] = None):
        """
        Generates and upgrades the helm chart

        :param options: A dictionary of command line arguments to pass to helm

        :Example:

        >>> self.upgrade_chart(options={"atomic": None, "version": "2.0"})
        """
        self.__check_if_installed()
        self.generate_chart()
        self.add_dependency_repos()
        update_depenedencies = "dependency-update"
        if options is not None and update_depenedencies in options:
            custom_check_output(
                f"helm dependency update {self.chart_folder_path.resolve()}"
            )
            del options[update_depenedencies]
        self.__handle_upgrade(options)

    @property
    def is_installed(self):
        """
        :return: True if chart with the given name is already installed in the chart \
            builders namespace, else False
        """
        installations = get_helm_installations(self.namespace)
        if not installations:
            return False
        return self.chart_info.name in installations["NAME"]



"""
Contains PriorityClass class
"""

from typing import Optional

from avionix.kube.base_objects import Scheduling
from avionix.kube.meta import ObjectMeta


class PriorityClass(Scheduling):
    """
    :param metadata: Standard object's metadata. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata  # noqa
    :param value: The value of this priority class. This is the actual priority that \
        pods receive when they have the name of this class in their pod spec.
    :param global_default: globalDefault specifies whether this PriorityClass should be \
        considered as the default priority for pods that do not have any priority \
        class. Only one PriorityClass can be marked as `globalDefault`. However, if \
        more than one PriorityClasses exists with their `globalDefault` field set to \
        true, the smallest value of such global default PriorityClasses will be used \
        as the default priority.
    :param description: description is an arbitrary string that usually provides \
        guidelines on when this priority class should be used.
    :param preemption_policy: PreemptionPolicy is the Policy for preempting pods with \
        lower priority. One of Never, PreemptLowerPriority. Defaults to \
        PreemptLowerPriority if unset. This field is alpha-level and is only honored \
        by servers that enable the NonPreemptingPriority feature.
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        value: int,
        global_default: Optional[bool] = None,
        description: Optional[str] = None,
        preemption_policy: Optional[str] = None,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.globalDefault = global_default
        self.value = value
        self.description = description
        self.preemptionPolicy = preemption_policy



from logging import error, info
from subprocess import STDOUT, CalledProcessError, check_output


def custom_check_output(command: str):
    info(f"Running command: {command}")
    try:
        output = check_output(command.split(" "), stderr=STDOUT).decode("utf-8")
    except CalledProcessError as err:
        error(err.output.decode("utf-8"))
        raise err
    info(f"Output from command:\n{output}")
    return output



"""
Classes related to scaling Kubernetes pods
"""

from typing import Optional

from avionix.kube.base_objects import Autoscaling
from avionix.kube.meta import ObjectMeta
from avionix.yaml.yaml_handling import HelmYaml


class CrossVersionObjectReference(Autoscaling):
    """
    :param name: Name of the referent; More info: \
        http://kubernetes.io/docs/user-guide/identifiers#names
    :param api_version: API version of the referent
    """

    def __init__(self, name: str, api_version: Optional[str] = None):
        super().__init__(api_version)
        self.name = name


class HorizontalPodAutoscalerSpec(HelmYaml):
    """
    :param max_replicas: upper limit for the number of pods that can be set by the \
        autoscaler; cannot be smaller than MinReplicas.
    :param scale_target_ref: reference to scaled resource; horizontal pod autoscaler \
        will learn the current resource consumption and will set the desired number of \
        pods by using its Scale subresource.
    :param min_replicas: minReplicas is the lower limit for the number of replicas to \
        which the autoscaler can scale down.  It defaults to 1 pod.  minReplicas is \
        allowed to be 0 if the alpha feature gate HPAScaleToZero is enabled and at \
        least one Object or External metric is configured.  Scaling is active as long \
        as at least one metric value is available.
    :param target_cpuutilization_percentage: target average CPU utilization \
        (represented as a percentage of requested CPU) over all the pods; if not \
        specified the default autoscaling policy will be used.
    """

    def __init__(
        self,
        max_replicas: int,
        scale_target_ref: CrossVersionObjectReference,
        min_replicas: Optional[int] = None,
        target_cpuutilization_percentage: Optional[int] = None,
    ):
        self.maxReplicas = max_replicas
        self.scaleTargetRef = scale_target_ref
        self.minReplicas = min_replicas
        self.targetCPUUtilizationPercentage = target_cpuutilization_percentage


class HorizontalPodAutoscaler(Autoscaling):
    """
    :param metadata: Standard object metadata. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata  # noqa
    :param spec: behaviour of autoscaler. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#spec-and-status.  # noqa
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        spec: HorizontalPodAutoscalerSpec,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec



"""
Various webhook configurations
"""

from typing import List, Optional

from avionix.kube.apiextensions import WebhookClientConfig
from avionix.kube.base_objects import AdmissionRegistration
from avionix.kube.meta import LabelSelector, ObjectMeta
from avionix.yaml.yaml_handling import HelmYaml


class RuleWithOperations(HelmYaml):
    """
    :param api_groups: APIGroups is the API groups the resources belong to. '\\*' is \
        all groups. If '\\*' is present, the length of the slice must be one. \
        Required.
    :param api_versions: APIVersions is the API versions the resources belong to. '\\*' \
        is all versions. If '\\*' is present, the length of the slice must be one. \
        Required.
    :param operations: Operations is the operations the admission hook cares about - \
        CREATE, UPDATE, or * for all operations. If '\\*' is present, the length of \
        the slice must be one. Required.
    :param resources: Resources is a list of resources this rule applies to.  For \
        example: 'pods' means pods. 'pods/log' means the log subresource of pods. \
        '\\*' means all resources, but not subresources. 'pods/\\*' means all \
        subresources of pods. '\\*/scale' means all scale subresources. '\\*/\\*' \
        means all resources and their subresources.  If wildcard is present, the \
        validation rule will ensure resources do not overlap with each other.  \
        Depending on the enclosing object, subresources might not be allowed. \
        Required.
    :param scope: scope specifies the scope of this rule. Valid values are "Cluster", \
        "Namespaced", and "*" "Cluster" means that only cluster-scoped resources will \
        match this rule. Namespace API objects are cluster-scoped. "Namespaced" means \
        that only namespaced resources will match this rule. "*" means that there are \
        no scope restrictions. Subresources match the scope of their parent resource. \
        Default is "*".
    """

    def __init__(
        self,
        api_groups: List[str],
        api_versions: List[str],
        operations: List[str],
        resources: List[str],
        scope: Optional[str] = None,
    ):
        self.apiGroups = api_groups
        self.apiVersions = api_versions
        self.operations = operations
        self.resources = resources
        self.scope = scope


class ValidatingWebhook(HelmYaml):
    """
    :param name: The name of the admission webhook. Name should be fully qualified, \
        e.g., imagepolicy.kubernetes.io, where "imagepolicy" is the name of the \
        webhook, and kubernetes.io is the name of the organization. Required.
    :param admission_review_versions: AdmissionReviewVersions is an ordered list of \
        preferred `AdmissionReview` versions the Webhook expects. API server will try \
        to use first version in the list which it supports. If none of the versions \
        specified in this list supported by API server, validation will fail for this \
        object. If a persisted webhook configuration specifies allowed versions and \
        does not include any versions known to the API Server, calls to the webhook \
        will fail and be subject to the failure policy.
    :param client_config: ClientConfig defines how to communicate with the hook. \
        Required
    :param side_effects: SideEffects states whether this webhook has side effects. \
        Acceptable values are: None, NoneOnDryRun (webhooks created via v1beta1 may \
        also specify Some or Unknown). Webhooks with side effects MUST implement a \
        reconciliation system, since a request may be rejected by a future step in the \
        admission change and the side effects therefore need to be undone. Requests \
        with the dryRun attribute will be auto-rejected if they match a webhook with \
        sideEffects == Unknown or Some.
    :param failure_policy: FailurePolicy defines how unrecognized errors from the \
        admission endpoint are handled - allowed values are Ignore or Fail. Defaults \
        to Fail.
    :param match_policy: matchPolicy defines how the "rules" list is used to match \
        incoming requests. Allowed values are "Exact" or "Equivalent".  - Exact: match \
        a request only if it exactly matches a specified rule. For example, if \
        deployments can be modified via apps/v1, apps/v1beta1, and extensions/v1beta1, \
        but "rules" only included `apiGroups:["apps"], apiVersions:["v1"], resources: \
        ["deployments"]`, a request to apps/v1beta1 or extensions/v1beta1 would not be \
        sent to the webhook.  - Equivalent: match a request if modifies a resource \
        listed in rules, even via another API group or version. For example, if \
        deployments can be modified via apps/v1, apps/v1beta1, and extensions/v1beta1, \
        and "rules" only included `apiGroups:["apps"], apiVersions:["v1"], resources: \
        ["deployments"]`, a request to apps/v1beta1 or extensions/v1beta1 would be \
        converted to apps/v1 and sent to the webhook.  Defaults to "Equivalent"
    :param namespace_selector: NamespaceSelector decides whether to run the webhook on \
        an object based on whether the namespace for that object matches the selector. \
        If the object itself is a namespace, the matching is performed on \
        object.metadata.labels. If the object is another cluster scoped resource, it \
        never skips the webhook.  For example, to run the webhook on any objects whose \
        namespace is not associated with "runlevel" of "0" or "1";  you will set the \
        selector as follows: "namespaceSelector": {   "matchExpressions": [     {      \
         "key": "runlevel",       "operator": "NotIn",       "values": [         "0",  \
               "1"       ]     }   ] }  If instead you want to only run the webhook on \
        any objects whose namespace is associated with the "environment" of "prod" or \
        "staging"; you will set the selector as follows: "namespaceSelector": {   \
        "matchExpressions": [     {       "key": "environment",       "operator": \
        "In",       "values": [         "prod",         "staging"       ]     }   ] }  \
        See https://kubernetes.io/docs/concepts/overview/working-with-objects/labels \
        for more examples of label selectors.  Default to the empty LabelSelector, \
        which matches everything.
    :param object_selector: ObjectSelector decides whether to run the webhook based on \
        if the object has matching labels. objectSelector is evaluated against both \
        the oldObject and newObject that would be sent to the webhook, and is \
        considered to match if either object matches the selector. A null object \
        (oldObject in the case of create, or newObject in the case of delete) or an \
        object that cannot have labels (like a DeploymentRollback or a PodProxyOptions \
        object) is not considered to match. Use the object selector only if the \
        webhook is opt-in, because end users may skip the admission webhook by setting \
        the labels. Default to the empty LabelSelector, which matches everything.
    :param rules: Rules describes what operations on what resources/subresources the \
        webhook cares about. The webhook cares about an operation if it matches _any_ \
        Rule. However, in order to prevent ValidatingAdmissionWebhooks and \
        MutatingAdmissionWebhooks from putting the cluster in a state which cannot be \
        recovered from without completely disabling the plugin, \
        ValidatingAdmissionWebhooks and MutatingAdmissionWebhooks are never called on \
        admission requests for ValidatingWebhookConfiguration and \
        MutatingWebhookConfiguration objects.
    :param timeout_seconds: TimeoutSeconds specifies the timeout for this webhook. \
        After the timeout passes, the webhook call will be ignored or the API call \
        will fail based on the failure policy. The timeout value must be between 1 and \
        30 seconds. Default to 10 seconds.
    """

    def __init__(
        self,
        name: str,
        admission_review_versions: List[str],
        client_config: WebhookClientConfig,
        side_effects: str,
        failure_policy: Optional[str] = None,
        match_policy: Optional[str] = None,
        namespace_selector: Optional[LabelSelector] = None,
        object_selector: Optional[LabelSelector] = None,
        rules: Optional[List[RuleWithOperations]] = None,
        timeout_seconds: Optional[int] = None,
    ):
        self.name = name
        self.admissionReviewVersions = admission_review_versions
        self.clientConfig = client_config
        self.sideEffects = side_effects
        self.failurePolicy = failure_policy
        self.matchPolicy = match_policy
        self.namespaceSelector = namespace_selector
        self.objectSelector = object_selector
        self.rules = rules
        self.timeoutSeconds = timeout_seconds


class ValidatingWebhookConfiguration(AdmissionRegistration):
    """
    :param metadata: Standard object metadata; More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata.  # noqa
    :param webhooks: Webhooks is a list of webhooks and the affected resources and \
        operations.
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        webhooks: List[ValidatingWebhook],
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.webhooks = webhooks


class MutatingWebhook(HelmYaml):
    """
    :param name: The name of the admission webhook. Name should be fully qualified, \
        e.g., imagepolicy.kubernetes.io, where "imagepolicy" is the name of the \
        webhook, and kubernetes.io is the name of the organization. Required.
    :param admission_review_versions: AdmissionReviewVersions is an ordered list of \
        preferred `AdmissionReview` versions the Webhook expects. API server will try \
        to use first version in the list which it supports. If none of the versions \
        specified in this list supported by API server, validation will fail for this \
        object. If a persisted webhook configuration specifies allowed versions and \
        does not include any versions known to the API Server, calls to the webhook \
        will fail and be subject to the failure policy.
    :param client_config: ClientConfig defines how to communicate with the hook. \
        Required
    :param side_effects: SideEffects states whether this webhook has side effects. \
        Acceptable values are: None, NoneOnDryRun (webhooks created via v1beta1 may \
        also specify Some or Unknown). Webhooks with side effects MUST implement a \
        reconciliation system, since a request may be rejected by a future step in the \
        admission change and the side effects therefore need to be undone. Requests \
        with the dryRun attribute will be auto-rejected if they match a webhook with \
        sideEffects == Unknown or Some.
    :param failure_policy: FailurePolicy defines how unrecognized errors from the \
        admission endpoint are handled - allowed values are Ignore or Fail. Defaults \
        to Fail.
    :param match_policy: matchPolicy defines how the "rules" list is used to match \
        incoming requests. Allowed values are "Exact" or "Equivalent".  - Exact: match \
        a request only if it exactly matches a specified rule. For example, if \
        deployments can be modified via apps/v1, apps/v1beta1, and extensions/v1beta1, \
        but "rules" only included `apiGroups:["apps"], apiVersions:["v1"], resources: \
        ["deployments"]`, a request to apps/v1beta1 or extensions/v1beta1 would not be \
        sent to the webhook.  - Equivalent: match a request if modifies a resource \
        listed in rules, even via another API group or version. For example, if \
        deployments can be modified via apps/v1, apps/v1beta1, and extensions/v1beta1, \
        and "rules" only included `apiGroups:["apps"], apiVersions:["v1"], resources: \
        ["deployments"]`, a request to apps/v1beta1 or extensions/v1beta1 would be \
        converted to apps/v1 and sent to the webhook.  Defaults to "Equivalent"
    :param namespace_selector: NamespaceSelector decides whether to run the webhook on \
        an object based on whether the namespace for that object matches the selector. \
        If the object itself is a namespace, the matching is performed on \
        object.metadata.labels. If the object is another cluster scoped resource, it \
        never skips the webhook.  For example, to run the webhook on any objects whose \
        namespace is not associated with "runlevel" of "0" or "1";  you will set the \
        selector as follows: "namespaceSelector": {   "matchExpressions": [     {      \
         "key": "runlevel",       "operator": "NotIn",       "values": [         "0",  \
               "1"       ]     }   ] }  If instead you want to only run the webhook on \
        any objects whose namespace is associated with the "environment" of "prod" or \
        "staging"; you will set the selector as follows: "namespaceSelector": {   \
        "matchExpressions": [     {       "key": "environment",       "operator": \
        "In",       "values": [         "prod",         "staging"       ]     }   ] }  \
        See https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/ \
        for more examples of label selectors.  Default to the empty LabelSelector, \
        which matches everything.
    :param object_selector: ObjectSelector decides whether to run the webhook based on \
        if the object has matching labels. objectSelector is evaluated against both \
        the oldObject and newObject that would be sent to the webhook, and is \
        considered to match if either object matches the selector. A null object \
        (oldObject in the case of create, or newObject in the case of delete) or an \
        object that cannot have labels (like a DeploymentRollback or a PodProxyOptions \
        object) is not considered to match. Use the object selector only if the \
        webhook is opt-in, because end users may skip the admission webhook by setting \
        the labels. Default to the empty LabelSelector, which matches everything.
    :param reinvocation_policy: reinvocationPolicy indicates whether this webhook \
        should be called multiple times as part of a single admission evaluation. \
        Allowed values are "Never" and "IfNeeded".  Never: the webhook will not be \
        called more than once in a single admission evaluation.  IfNeeded: the webhook \
        will be called at least one additional time as part of the admission \
        evaluation if the object being admitted is modified by other admission plugins \
        after the initial webhook call. Webhooks that specify this option *must* be \
        idempotent, able to process objects they previously admitted. Note: * the \
        number of additional invocations is not guaranteed to be exactly one. * if \
        additional invocations result in further modifications to the object, webhooks \
        are not guaranteed to be invoked again. * webhooks that use this option may be \
        reordered to minimize the number of additional invocations. * to validate an \
        object after all mutations are guaranteed complete, use a validating admission \
        webhook instead.  Defaults to "Never".
    :param rules: Rules describes what operations on what resources/subresources the \
        webhook cares about. The webhook cares about an operation if it matches _any_ \
        Rule. However, in order to prevent ValidatingAdmissionWebhooks and \
        MutatingAdmissionWebhooks from putting the cluster in a state which cannot be \
        recovered from without completely disabling the plugin, \
        ValidatingAdmissionWebhooks and MutatingAdmissionWebhooks are never called on \
        admission requests for ValidatingWebhookConfiguration and \
        MutatingWebhookConfiguration objects.
    :param timeout_seconds: TimeoutSeconds specifies the timeout for this webhook. \
        After the timeout passes, the webhook call will be ignored or the API call \
        will fail based on the failure policy. The timeout value must be between 1 and \
        30 seconds. Default to 10 seconds.
    """

    def __init__(
        self,
        name: str,
        admission_review_versions: List[str],
        client_config: WebhookClientConfig,
        side_effects: str,
        failure_policy: Optional[str] = None,
        match_policy: Optional[str] = None,
        namespace_selector: Optional[LabelSelector] = None,
        object_selector: Optional[LabelSelector] = None,
        reinvocation_policy: Optional[str] = None,
        rules: Optional[List[RuleWithOperations]] = None,
        timeout_seconds: Optional[int] = None,
    ):
        self.name = name
        self.admissionReviewVersions = admission_review_versions
        self.clientConfig = client_config
        self.sideEffects = side_effects
        self.failurePolicy = failure_policy
        self.matchPolicy = match_policy
        self.namespaceSelector = namespace_selector
        self.objectSelector = object_selector
        self.reinvocationPolicy = reinvocation_policy
        self.rules = rules
        self.timeoutSeconds = timeout_seconds


class MutatingWebhookConfiguration(AdmissionRegistration):
    """
    :param metadata: Standard object metadata; More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#metadata.  # noqa
    :param webhooks: Webhooks is a list of webhooks and the affected resources and \
        operations.
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        webhooks: List[MutatingWebhook],
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.webhooks = webhooks



"""
Classes making up the main interfaces for other Kubernetes classes
"""

from typing import Optional

from avionix.options import DEFAULTS
from avionix.yaml.yaml_handling import HelmYaml


class KubernetesBaseObject(HelmYaml):
    """
    Base object for other kubernetes objects to inherit from
    Required fields come from
    https://kubernetes.io/docs/concepts/overview/working-with-objects/kubernetes-objects/
    """

    _version_prefix = ""
    _base_object_name = "KubernetesBaseObject"
    _non_standard_version = ""

    def __init__(
        self,
        api_version: Optional[str] = None,
        kind: Optional[str] = None,
        metadata=None,
    ):
        if kind is None:
            self.kind = self.__get_kube_object_type().__name__
        else:
            self.kind = kind

        self.apiVersion = self._get_api_version(api_version)

        self.metadata = metadata

    def _get_api_version(self, api_version: Optional[str]):
        if self._non_standard_version:
            return self._version_prefix + self._non_standard_version
        if api_version is None:
            return self._version_prefix + DEFAULTS["default_api_version"]
        return api_version

    def __get_kube_object_type(self):
        # Get all inherited to find classes exact kube object
        mro = type(self).__mro__
        for i, class_ in enumerate(mro):
            if class_.__name__ == type(self)._base_object_name:
                return mro[i - 2]
        raise Exception("KubernetesObject ancestor class not found!")


class Core(KubernetesBaseObject):
    """
    Base object for other kubernetes objects to inherit from
    Required fields come from
    https://kubernetes.io/docs/concepts/overview/working-with-objects/kubernetes-objects/
    """


class Apps(KubernetesBaseObject):
    """
    Base class for apps group
    """

    _version_prefix = "apps/"


class AdmissionRegistration(KubernetesBaseObject):
    """
    Base class for admission registration group
    """

    _version_prefix = "admissionregistration.k8s.io/"


class ApiExtensions(KubernetesBaseObject):
    """
    Base class for api extensions group
    """

    _version_prefix = "apiextensions.k8s.io/"


class ApiRegistration(KubernetesBaseObject):
    """
    Base class for api registration
    """

    _version_prefix = "apiregistration.k8s.io/"


class Extensions(KubernetesBaseObject):
    """
    Base class for api registration
    """

    _version_prefix = "extensions/"


class Batch(KubernetesBaseObject):
    """
    Base class for api registration
    """

    _version_prefix = "batch/"


class RbacAuthorization(KubernetesBaseObject):
    """
    Base class for rbac authorization
    """

    _version_prefix = "rbac.authorization.k8s.io/"


class Storage(KubernetesBaseObject):

    _version_prefix = "storage.k8s.io/"


class Authentication(KubernetesBaseObject):

    _version_prefix = "authentication.k8s.io/"


class Authorization(KubernetesBaseObject):

    _version_prefix = "authorization.k8s.io/"


class Autoscaling(KubernetesBaseObject):

    _version_prefix = "autoscaling/"


class Coordination(KubernetesBaseObject):

    _version_prefix = "coordination.k8s.io/"


class Networking(KubernetesBaseObject):

    _version_prefix = "networking.k8s.io/"


class Node(KubernetesBaseObject):

    _version_prefix = "node.k8s.io/"


class Scheduling(KubernetesBaseObject):

    _version_prefix = "scheduling.k8s.io/"


class Policy(KubernetesBaseObject):

    _version_prefix = "policy/"


class Certificates(KubernetesBaseObject):

    _version_prefix = "certificates.k8s.io/"


class Discovery(KubernetesBaseObject):

    _version_prefix = "discovery.k8s.io/"


class Meta(KubernetesBaseObject):

    _version_prefix = "meta.k8s.io/"



from copy import deepcopy
from datetime import datetime
import re
from typing import Optional, Union

from yaml import dump


def is_empty_yaml(value):
    # If value is None, [], {} do not include value
    return not value and not isinstance(value, (bool, str))


def is_private_var(key: str):
    if re.match(r"_+.*", key):
        return True
    return False


class HelmYaml:
    def __clean_nested(self, dictionary_or_list: Union[dict, list]):
        if isinstance(dictionary_or_list, list):
            cleaned_list = []
            for value in dictionary_or_list:
                if is_empty_yaml(value):
                    continue

                if isinstance(value, (dict, list)):
                    cleaned_dict_or_list = self.__clean_nested(value)

                    if cleaned_dict_or_list:
                        cleaned_list.append(self.__clean_nested(value))

                elif isinstance(value, HelmYaml):
                    cleaned_dict_or_list = value.to_dict()

                    if cleaned_dict_or_list:
                        cleaned_list.append(cleaned_dict_or_list)

                else:
                    cleaned_list.append(value)
            return cleaned_list

        elif isinstance(dictionary_or_list, dict):
            cleaned_dict = {}
            for key, value in dictionary_or_list.items():
                if is_private_var(key):
                    continue

                if is_empty_yaml(value):
                    continue

                elif isinstance(value, (dict, list)):
                    cleaned_dict_or_list = self.__clean_nested(value)

                    if cleaned_dict_or_list:
                        cleaned_dict[key] = cleaned_dict_or_list

                elif isinstance(value, HelmYaml):
                    cleaned_dict_or_list = value.to_dict()

                    if cleaned_dict_or_list:
                        cleaned_dict[key] = cleaned_dict_or_list
                else:
                    cleaned_dict[key] = value
            return cleaned_dict

    def __str__(self):
        return dump(self.to_dict())

    def to_dict(self):
        dictionary = deepcopy(self.__dict__)
        return self.__clean_nested(dictionary)

    @staticmethod
    def _get_kube_date_string(datetime_obj: Optional[datetime]):
        return (
            datetime_obj.strftime("%Y-%m-%dT%H:%M:%S.%fZ%Z")
            if datetime_obj
            else datetime_obj
        )

    def __setitem__(self, key, value):
        self.__dict__[key] = value



"""
Classes for token based authentication
"""

from typing import List, Optional

from avionix.kube.base_objects import Authentication
from avionix.kube.meta import ObjectMeta
from avionix.yaml.yaml_handling import HelmYaml


class BoundObjectReference(HelmYaml):
    """
    :param api_version: API version of the referent.
    :param kind: Kind of the referent. Valid kinds are 'Pod' and 'Secret'.
    :param name: Name of the referent.
    :param uid: UID of the referent.
    """

    def __init__(
        self, api_version: str, kind: str, name: str, uid: Optional[str] = None
    ):
        self.apiVersion = api_version
        self.kind = kind
        self.name = name
        self.uid = uid


class TokenRequestSpec(HelmYaml):
    """
    :param audiences: Audiences are the intendend audiences of the token. A recipient \
        of a token must identitfy themself with an identifier in the list of audiences \
        of the token, and otherwise should reject the token. A token issued for \
        multiple audiences may be used to authenticate against any of the audiences \
        listed but implies a high degree of trust between the target audiences.
    :param bound_object_ref: BoundObjectRef is a reference to an object that the token \
        will be bound to. The token will only be valid for as long as the bound object \
        exists. NOTE: The API server's TokenReview endpoint will validate the \
        BoundObjectRef, but other audiences may not. Keep ExpirationSeconds small if \
        you want prompt revocation.
    :param expiration_seconds: ExpirationSeconds is the requested duration of validity \
        of the request. The token issuer may return a token with a different validity \
        duration so a client needs to check the 'expiration' field in a response.
    """

    def __init__(
        self,
        audiences: List[str],
        bound_object_ref: BoundObjectReference,
        expiration_seconds: int,
    ):
        self.audiences = audiences
        self.boundObjectRef = bound_object_ref
        self.expirationSeconds = expiration_seconds


class TokenRequest(Authentication):
    """
    :param metadata: None
    :param spec: None
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        spec: TokenRequestSpec,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec


class TokenReviewSpec(HelmYaml):
    """
    :param audiences: Audiences is a list of the identifiers that the resource server \
        presented with the token identifies as. Audience-aware token authenticators \
        will verify that the token was intended for at least one of the audiences in \
        this list. If no audiences are provided, the audience will default to the \
        audience of the Kubernetes apiserver.
    :param token: Token is the opaque bearer token.
    """

    def __init__(self, audiences: List[str], token: str):
        self.audiences = audiences
        self.token = token


class TokenReview(Authentication):
    """
    :param metadata: None
    :param spec: Spec holds information about the request being evaluated
    :param api_version: APIVersion defines the versioned schema of this representation \
        of an object. Servers should convert recognized schemas to the latest internal \
        value, and may reject unrecognized values. More info: \
        https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources  # noqa
    """

    def __init__(
        self,
        metadata: ObjectMeta,
        spec: TokenReviewSpec,
        api_version: Optional[str] = None,
    ):
        super().__init__(api_version)
        self.metadata = metadata
        self.spec = spec
